from __future__ import print_function
import os
import sys
import json
import subprocess
from Accuinsight.modeler.core.LcConst import LcConst

_BASH_RC_PATH = '/home/notebook/.bashrc'


def get_os_env(env_type='LC', env_path=None, env_file=None):
    _ENV_PREFIX = env_type + '_'

    env_value = {}
    if env_path is None:
        env_path = '/home/work'
    if env_file is None:
        env_file = '.env'

    env_file = open(env_path + '/' + env_file, 'r')
    while True:
        line = env_file.readline()
        if not line:
            break
        if line:
            if _ENV_PREFIX in line:
                key, value = line.split('=')
                if key is not None:
                    if _ENV_PREFIX in key:
                        value = value.rstrip()
                        env_value.setdefault(key, value)

    return env_value


def is_in_ipython():
    ip = False
    if 'ipykernel' in sys.modules:
        ip = True
    return ip
    
    
def __list_jupyterlab_workspace(targtet_path):
    file_lists = []
    for (dirpath, dirnames, filenames) in os.walk(targtet_path):
        for fn in filenames:
            if not fn.endswith('.swp'):
                if 'labworkspaces' in fn: continue
                file_lists.append(fn)
            break
    return file_lists


def __get_current_file(target_path, target_file):
    with open(target_path + '/' + target_file, 'r') as tf:
        strlines = tf.readlines()
        aaa = json.loads(strlines[0])
        current_notebook = aaa['data']['layout-restorer:data']['main']['current']
    return current_notebook


def __install_package(package_name):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package_name])


def get_current_notebook():
    current_notebook = ''
    is_jupyter = is_in_ipython()
    if is_jupyter == True:
        target_path = LcConst.ENV_JUPYTER_WORKSPACE
        # '/home/notebook/.jupyter/lab/workspaces'
        if os.path.isdir(target_path) is True:
            ret_file = __list_jupyterlab_workspace(targtet_path=target_path)[0]
            current_notebook = __get_current_file(target_path=target_path, target_file=ret_file)
            current_notebook = current_notebook.split(':')[1]
        else:
            __install_package('ipyparams')
            import ipyparams
            current_notebook = ipyparams.raw_url
            current_notebook = current_notebook.split('notebooks/')[1]
    return current_notebook



import os
import sys
import inspect
import json
import logging
import warnings
import time
import datetime
from collections import OrderedDict

import boto3
from hdfs3 import HDFileSystem
import gorilla
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
import numpy as np
import tensorflow
from tensorflow.keras.callbacks import CSVLogger
from Accuinsight.modeler.core import func, path, get
from Accuinsight.modeler.core.func import get_time
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.modeler.core.LcConst.LcConst import RUN_NAME_TENSORFLOW, ENV_JUPYTER_HOME_DIR
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import set_current_runs, clear_runs, set_model_json_path, \
    set_visual_csv_path, set_visual_json_path, set_best_model_json_path, set_best_model_h5_path, print_run_info, \
    set_python_dependencies, set_run_name, set_model_file_path, set_prefix_path, set_shap_json_path
from Accuinsight.modeler.core.get_for_visual import roc_pr_curve, get_true_y, get_visual_info_regressor
from Accuinsight.modeler.clients.modeler_api import LifecycleRestApi
from Accuinsight.modeler.utils.dependency.dependencies import gather_sources_and_dependencies
from Accuinsight.modeler.utils.dl_utils import delete_files_except_best, get_best_model_path
from Accuinsight.modeler.utils.os_getenv import is_in_ipython, get_current_notebook, get_os_env
from Accuinsight.modeler.utils.runs_utils import get_aws_info, ProgressPercentage
from Accuinsight.modeler.core.feature_contribution import shap_value

logging.basicConfig(level=logging.INFO,
                    format='%(message)s')

warnings.filterwarnings("ignore")


class accuinsight(object):
    def __init__(self):
        self.StorageInfo = None
        self.target_name = None
        self.save_path = None
        self.endpoint = None
        self.thresholds = None
        self.token = None
        self.channel_id = None
        self.message = None
        self.feature_name = None
        self.storage_info_from_set = OrderedDict()
        
    def set_storage(self,
                    access_key=None,     # aws-s3
                    secret_key=None,     # aws-s3
                    region=None,         # aws-s3
                    bucket_name=None,    # aws-s3
                    hdfs_uri=None,       # pipeline-hdfs
                    file_path=None,      # common
                    target=None,         # common
                    save_json=False,     # common - if it is TRUE, saving a storage information as a json file.
                    save_path=None       # common - if save_path is not None, the json file is saved in this path.
    ):
        
        if target is None:
            raise ValueError('target을 입력해주시기 바랍니다.')
        
        ## make a directory for saving json file
        if save_json:
            save_prefix_1 = os.path.join(ENV_JUPYTER_HOME_DIR, 'runs')
            if not os.path.isdir(save_prefix_1):
                try:
                    os.mkdir(save_prefix_1)
                except:
                    pass
            save_prefix_2 = os.path.join(save_prefix_1, 'storage-info-json')
            if not os.path.isdir(save_prefix_2):
                try:
                    os.mkdir(save_prefix_2)
                except:
                    pass
        
        ### HDFS
        if hdfs_uri is not None:
            storage_type = 'hdfs'
            
            split_1 = hdfs_uri.split('//')[1].split('/')
            split_2 = [i.split(':') for i in split_1 if ':' in i][0]
            
            self.storage_info_from_set['host'] = split_2[0]
            self.storage_info_from_set['port'] = split_2[1]
            self.storage_info_from_set['filePath'] = os.path.join('/', *split_1[1:])
            self.storage_info_from_set['target'] = target
        
        ### AWS - S3
        elif bucket_name is not None:
            storage_type = 's3'
            
            self.storage_info_from_set['myAccessKey'] = access_key
            self.storage_info_from_set['mySecretKey'] = secret_key
            self.storage_info_from_set['endpoint'] = 's3.' + region + '.amazonaws.com'
            self.storage_info_from_set['region'] = region
            self.storage_info_from_set['bucketType'] = 's3'
            self.storage_info_from_set['bucketName'] = bucket_name
            self.storage_info_from_set['filePath'] = file_path
            self.storage_info_from_set['fileName'] = file_path.split('/')[-1]
            self.storage_info_from_set['target'] = target

        ### Local
        else:
            self.storage_info_from_set['filePath'] = file_path
            self.storage_info_from_set['fileName'] = file_path.split('/')[-1]
            self.storage_info_from_set['target'] = target
        
        if save_json:
            if save_path is None:
                timestamp = str(datetime.datetime.now().date()).replace('-', '')
                dir_list = os.listdir(save_prefix_2)
                num = len([True for i in dir_list if timestamp in i])
                save_path = timestamp + '-' + str(num+1).zfill(2) + '-' + storage_type + '-connection-info' + '.json'
                
            else:    # save_path is not None
                pass
                
            storage_info_full_path = os.path.join(save_prefix_2, save_path)
            with open(storage_info_full_path, 'w', encoding='utf-8') as save_file:
                json.dump(self.storage_info_from_set, save_file, indent='\t')
                
        if not save_json and save_path is not None:
            raise ValueError("'save_json=True'로 설정하시기 바랍니다.")

    def get_file(self, storage_json_file_name=None):
        if storage_json_file_name is not None:
        
            # meta file 저장하는 폴더 만들면, 이 경로는 수정돼야 함.
            storage_info_json_path = os.path.join(ENV_JUPYTER_HOME_DIR, storage_json_file_name)
       
            with open(storage_info_json_path) as jsonFile:
                self.StorageInfo = json.load(jsonFile)
        
        else:   # storage_json_file_name == None, when using set_storage() method.
            self.StorageInfo = self.storage_info_from_set
            
        ## define the target name
        self.target_name = self.StorageInfo['target']
            
        ### HDFS
        if 'host' in self.StorageInfo.keys():
            
            ## path for saving data
            save_dir = os.path.join(ENV_JUPYTER_HOME_DIR, 'data_from_hdfs')
            if os.path.exists(save_dir) == False:
                os.mkdir(save_dir)
            else:
                pass
            
            split_list = list(filter(None, self.StorageInfo['filePath'].split('/')))
            
            HOST = self.StorageInfo['host']
            PORT = int(self.StorageInfo['port'])
            FILE_PATH = self.StorageInfo['filePath']
            FILE_DIR = '/'
            for i in split_list[:len(split_list)-1]:
                FILE_DIR = os.path.join(FILE_DIR, i)
            FILE_NAME = split_list[-1]
            
            self.save_file_name = FILE_NAME.split('.')[0] + '_' + str(datetime.datetime.now().date()).replace('-', '') + '.' + FILE_NAME.split('.')[1]
            self.save_path = os.path.join(save_dir, self.save_file_name)
            self.endpoint = os.path.join('hdfs://', HOST, self.StorageInfo['port']) + FILE_PATH
            
            hdfs = HDFileSystem(host = HOST, port = PORT)
            hdfs.ls(FILE_DIR)

            sys.stdout.write('%s %s %s' % ('Downloading file...', FILE_NAME, '\n'))
            time.sleep(1)
            hdfs.get(FILE_PATH, self.save_path)
            logging.info(self.save_path)
        
        ### AWS-S3
        elif 'bucketName' in self.StorageInfo.keys():
            BUCKET_TYPE = self.StorageInfo['bucketType']
            BUCKET_NAME = self.StorageInfo['bucketName']
            FILE_PATH = self.StorageInfo['filePath']
            FILE_NAME = self.StorageInfo['fileName']
            #FILE_TYPE = self.StorageInfo['fileType']
            #FILE_DELIM = self.StorageInfo['fileDelim']
            ACCESS_KEY = self.StorageInfo['myAccessKey']
            SECRET_KEY = self.StorageInfo['mySecretKey']
            REGION = self.StorageInfo['region']
            URL = self.StorageInfo['endpoint']

            ## path for saving data
            save_dir = os.path.join(ENV_JUPYTER_HOME_DIR, 'data_from_aws')
            if os.path.exists(save_dir) == False:
                os.mkdir(save_dir)
            else:
                pass
        
            self.save_file_name = FILE_NAME.split('.')[0] + '_' + str(datetime.datetime.now().date()).replace('-', '') + '.' + FILE_NAME.split('.')[1]
            self.save_path = os.path.join(save_dir, self.save_file_name)

            # endpoint
            pre_url = 'https://' + BUCKET_NAME + '.' + URL
            self.endpoint = os.path.join(pre_url, FILE_PATH)

            client = boto3.client(BUCKET_TYPE,
                                  aws_access_key_id=ACCESS_KEY,
                                  aws_secret_access_key=SECRET_KEY,
                                  region_name=REGION)

            transfer = boto3.s3.transfer.S3Transfer(client)

            progress = ProgressPercentage(client, BUCKET_NAME, FILE_PATH)

            sys.stdout.write('%s %s %s' % ('Downloading file...', FILE_NAME, '\n'))
            transfer.download_file(BUCKET_NAME, FILE_PATH, self.save_path, callback=progress)
            logging.info(self.save_path)

        ### Local
        else:
            self.save_path = self.StorageInfo['filePath']
    
    def set_slack(self, token=None, channel_id=None):
        self.token = token
        self.channel_id = channel_id

    def send_message(self, message=None, thresholds=None):
        if message is not None and thresholds is not None:
            raise ValueError("'message'와 'thresholds' 두 개의 arguments를 동시에 입력할 수 없습니다.")
        else:
            self.message = message
            self.thresholds = thresholds

    def set_features(self, feature_names):
        
        if str(type(feature_names)) == "<class 'list'>":
            self.feature_name = feature_names
        else:
            try:
                self.feature_name = feature_names.tolist()
            except:
                pass
            
            try:
                self.feature_name = list(feature_names)
            except AttributeError:
                raise ValueError('입력값으로 list 또는 pandas.core.indexes.base.Index를 사용하시기 바랍니다.')

    def autolog(self, tag=None, best_weights=False, model_monitor=False):
        global description, endpoint, var_model_file_path, message, thresholds, token, channel_id, best_weights_on, run_id, shap_on, feature_name
        description = tag
        endpoint = self.endpoint
        message = self.message
        thresholds = self.thresholds
        token = self.token
        channel_id = self.channel_id
    
        if best_weights:
            best_weights_on = True
        else:
            best_weights_on = False
        if model_monitor:
            shap_on = True
            if self.feature_name is None:
                try:
                    feature_name = get.feature_name(self.save_path, self.StorageInfo, self.target_name)

                except:
                    pass      # when user did not use <get_file> or <set_feature> function
            else:
                feature_name = self.feature_name
        else:
            shap_on = False
        run_id = None

        is_notebook = is_in_ipython()
        if is_notebook:
            var_model_file_path = get_current_notebook()
            _caller_globals = inspect.stack()[1][0].f_globals
            (
                mainfile,
                sources,
                dependencies
            ) = gather_sources_and_dependencies(
                globs=_caller_globals,
                save_git_info=False
            )
        else:
            _caller_globals = inspect.stack()[1][0].f_globals
            (
                mainfile,
                sources,
                dependencies
            ) = gather_sources_and_dependencies(
                globs=_caller_globals,
                save_git_info=True
            )
            var_model_file_path = mainfile['filename']

        class TrainHistoryCallbacks(tensorflow.keras.callbacks.Callback):
            def __init__(self, verbose=1, mode='auto'):
                super(TrainHistoryCallbacks, self).__init__()
                self.verbose = verbose
                self.best_epochs = 0
                self.epochs_since_last_save = 0
                self.mode = mode

            def on_train_begin(self, logs={}):
                if best_weights_on and shap_on:
                    logging.info('Using autolog(best_weights=True, model_monitor=True)')
                elif best_weights_on == True and shap_on == False:
                    logging.info('Using autolog(best_weights=True, model_monitor=False)')
                elif best_weights_on == False and shap_on == True:
                    logging.info('Using autolog(best_weights=False, model_monitor=True)')
                else:
                    logging.info('Using autolog(best_weights=False, model_monitor=False)')
                
                global start
                start = get_time.now()
                opt = self.model.optimizer.get_config()
                opt_key = list(opt.keys())[1:]
                opt_result = {k: np.float64(opt[k]) for k in opt_key}
                self.model_summary = OrderedDict()
                self.model_summary['data_version'] = endpoint
                self.model_summary['model_description'] = description
                self.model_summary['logging_time'] = get_time.logging_time()
                self.model_summary['run_id'] = func.get_run_id()
                self.model_summary['model_type'] = get.model_type(self.model)

                if hasattr(self.model.loss, 'get_config'):
                    self.model_summary['loss_function'] = self.model.loss.get_config()['name']
                else:
                    self.model_summary['loss_function'] = self.model.loss

                self.model_summary['optimizer_info'] = {opt['name']: opt_result}

                '''[get best model] on_train_begin '''
                self.best_weights = self.model.get_weights()

                self.dict_path = path.get_file_path(self.model, usedFramework='tensorflow')

                set_prefix_path(self.dict_path[LcConst.RUN_PREFIX_PATH])

                set_run_name(self.model_summary['model_type'], self.model_summary['run_id'])
                set_python_dependencies(py_depenpency=dependencies)

#                print('dict_path: ', self.dict_path)         ######################
        
            '''[get best model] on_epoch_end '''
            def on_epoch_end(self, epoch, logs=None):
                logs = logs or {}
                if epoch == 0:
                    if len(self.model.metrics_names) == 1 and 'loss' in self.model.metrics_names:
                        self.monitor = 'val_loss'
                    elif len(self.model.metrics_names) >= 2:
                        self.monitor = 'val_' + self.model.metrics_names[1]
                    # set monitoring option
                    if self.mode not in ['auto', 'min', 'max']:
                        warnings.warn('GetBest mode %s is unknown, '
                                      'fallback to auto mode.' % (self.mode), RuntimeWarning)
                        self.mode = 'auto'
                    if self.mode == 'min':
                        self.monitor_op = np.less
                        self.best = np.Inf
                    elif self.mode == 'max':
                        self.monitor_op = np.greater
                        self.best = -np.Inf
                    else:
                        if 'acc' in self.monitor or 'f1' in self.monitor:
                            self.monitor_op = np.greater
                            self.best = -np.Inf
                        else:
                            self.monitor_op = np.less
                            self.best = np.Inf
                else:
                    pass
                
                # Using best_weights
                if best_weights_on:
                    
                    # update best_weights
                    self.epochs_since_last_save += 1
                    if self.epochs_since_last_save >= 1:
                        self.epochs_since_last_save = 0
                        current = logs.get(self.monitor)
                        if current is None:
                            warnings.warn('Can pick best model only with %s available, '
                                      'skipping.' % (self.monitor), RuntimeWarning)
                        else:
                            if self.monitor_op(current, self.best):
                                self.best = current
                                self.best_epochs = epoch + 1
                                self.best_weights = self.model.get_weights()
                            else:
                                pass

                    self.current_value = current
    
                # Not using best_weights
                else:
                    self.last_epoch_metric = logs.get(self.monitor)
                    self.best_epochs = epoch + 1
                    current = logs.get(self.monitor)
                    self.current_value = current

                # model save path
                run_id = self.model_summary['model_type'] + '-' + self.model_summary['run_id']
                common_path = self.dict_path['save_model_path'] + run_id + '-epoch-' + str(epoch + 1).zfill(5) + '-' + self.monitor + '-' + str(current).zfill(5)
                save_model_path = common_path + '.json'
                save_weights_path = common_path + '.h5'
                # model to JSON
                model_json = self.model.to_json()
                with open(save_model_path, "w") as json_file:
                    json_file.write(model_json)
                # weights to H5
                self.model.save_weights(save_weights_path)

            def on_train_end(self, logs={}):
                '''[get best model] on_train_end '''
                if best_weights_on:
                    if self.verbose > 0:
                        print('\nUsing epoch %05d with %s: %0.5f' % (self.best_epochs, self.monitor, self.best))
                    self.model.set_weights(self.best_weights)  # set best model's weights
                    
                    self.model_summary['selected_metrics'] = {self.monitor: np.float64(self.best)}
                else:
                    self.model_summary['selected_metrics'] = {self.monitor: np.float64(self.last_epoch_metric)}
    
          #      print('model_summary: ', self.model_summary)   ##############################
                end = get_time.now()
                self.model_summary['time_delta'] = str(end - start)

                # path for model_info.json
                self.path_for_setting_model_json = self.dict_path['model_json']
                set_model_json_path(self.path_for_setting_model_json)

                model_json_full_path = self.dict_path['model_json_full']

                with open(model_json_full_path, 'w', encoding='utf-8') as save_file:
                    json.dump(self.model_summary, save_file, indent="\t")

                delete_files_except_best(run_id=self.model_summary['run_id'], epochs=str(self.best_epochs),
                                         path=self.dict_path)

                path_for_setting_model_json = self.dict_path['save_model_dir'] + \
                                              get_best_model_path(run_id=self.model_summary['run_id'],
                                                                  path=self.dict_path)['json']
                path_for_setting_model_h5 = self.dict_path['save_model_dir'] + \
                                            get_best_model_path(run_id=self.model_summary['run_id'],
                                                                path=self.dict_path)['h5']
                set_best_model_json_path(path_for_setting_model_json)
                set_best_model_h5_path(path_for_setting_model_h5)

                start_ts = int(start.timestamp())
                end_ts = int(end.timestamp())
                delta_ts = end_ts - start_ts

                clear_runs(start_ts, end_ts, delta_ts)
                accuinsight._send_message(metric=self.monitor,
                                          current_value=self.current_value,
                                          message=message,
                                          thresholds=thresholds,
                                          token=token,
                                          channel_id=channel_id)
                env_value = get_os_env('ENV')
                modeler_rest = LifecycleRestApi(env_value[LcConst.BACK_END_API_URL],
                                                env_value[LcConst.BACK_END_API_PORT],
                                                env_value[LcConst.BACK_END_API_URI])
                modeler_rest.lc_create_run()
                accuinsight._off_autolog()

        class visualCallbacks(tensorflow.keras.callbacks.Callback):
            def __init__(self, x_validation=None, y_validation=None):
                self.x_val = x_validation
                self.y_val = y_validation

            def on_train_end(self, logs={}):

                self.dict_path = path.get_file_path(self.model, usedFramework='tensorflow')
                # path for visual.json
                path_for_setting_visual_json = self.dict_path['visual_json']
                visual_json_full_path = self.dict_path['visual_json_full']
                set_visual_json_path(path_for_setting_visual_json)

                # classification
                if get.is_classification(self.model):
                    visual_classification_json = roc_pr_curve(self.model, self.x_val, self.y_val)

                    with open(visual_json_full_path, 'w', encoding='utf-8') as save_file:
                        json.dump(visual_classification_json, save_file, indent="\t")

                # regression
                else:
                    visual_regression_json = OrderedDict()
                    visual_regression_json['True_y'] = get_true_y(self.y_val)
                    visual_regression_json['Predicted_y'] = get_visual_info_regressor(self.model, self.x_val)

                    with open(visual_json_full_path, 'w', encoding='utf-8') as save_file:
                        json.dump(visual_regression_json, save_file, indent="\t")

        class shapCallbacks(tensorflow.keras.callbacks.Callback):
            def __init__(self, trainX, feature_name, run_id, trigger = shap_on):
                self.trainX = trainX
                self.trigger = shap_on
                self.run_id = run_id
                self.feature_name_in_shap = feature_name

            def on_train_end(self, logs={}):
                if self.trigger:
                    self.shap_value = shap_value(self.model, self.trainX, self.feature_name_in_shap)
                    
#                     func.insertDB(self.shap_value, 2)  # 수정: 2 -> self.run_id
                    
                    self.dict_path = path.get_file_path(self.model, usedFramework='tensorflow')

                    # path for shap.json
                    shap_json_full_path = self.dict_path['shap_json_full']
                    set_shap_json_path(self.dict_path['shap_json'])
                    
                    with open(shap_json_full_path, 'w', encoding='utf-8') as save_file:
                        json.dump(self.shap_value, save_file, indent='\t')

                else:
                    pass
                    
        def run_and_log_function(self, original, x, y, kwargs):
            dict_path = path.get_file_path(self, usedFramework='tensorflow')
            path_for_setting_visual_csv = dict_path['visual_csv']
            visual_csv_full_path = dict_path['visual_csv_full']

            # set current run
            set_current_runs(RUN_NAME_TENSORFLOW)
            set_model_file_path(var_model_file_path)
            set_visual_csv_path(path_for_setting_visual_csv)

            csv_logger = CSVLogger(visual_csv_full_path, append=True, separator=';')
            
            # get train data(x) for computing shap value
            if 'x':
                x_train = x
            if shap_on:
                get_shap = shapCallbacks(x_train, feature_name, run_id, trigger = shap_on)
            else:
                pass
            
            ''' save json for visualization '''
            # using validation_data argument
            if 'validation_data' in kwargs:
                validation_set = kwargs['validation_data']

                try:
                    x_val = validation_set[0]
                    y_val = validation_set[1]

                except:
                    iterator = iter(validation_set)
                    valid_set = next(iterator)
                    x_val = valid_set[0].numpy()
                    y_val = valid_set[1].numpy()

                get_visual = visualCallbacks(x_validation=x_val, y_validation=y_val)

            else:
                raise ValueError('"validation_data" does not exist.')

            if 'callbacks' in kwargs:
                kwargs['callbacks'] += [csv_logger]
                kwargs['callbacks'] += [get_visual]
                if shap_on:
                    kwargs['callbacks'] += [get_shap]
                else:
                    pass
                kwargs['callbacks'] += [TrainHistoryCallbacks()]
                

            else:
                kwargs['callbacks'] = [csv_logger]
                kwargs['callbacks'] += [get_visual]
                if shap_on:
                    kwargs['callbacks'] += [get_shap]
                else:
                    pass
                kwargs['callbacks'] += [TrainHistoryCallbacks()]

            return original(self, x, y, **kwargs)

        @gorilla.patch(tensorflow.keras.Model)
        def fit(self, x, y, **kwargs):
            original = gorilla.get_original_attribute(tensorflow.keras.Model, 'fit')
            unlogged_params = ['self', 'callbacks', 'validation_data', 'verbose']
            return run_and_log_function(self, original, x, y, kwargs)

        settings = gorilla.Settings(allow_hit=True, store_hit=True)
        gorilla.apply(gorilla.Patch(tensorflow.keras.Model, 'fit', fit, settings=settings))

    @staticmethod
    def _send_message(metric=None, current_value=None, message=None, thresholds=None, token=None, channel_id=None):
        logger = logging.getLogger(__name__)
        
        if token is not None and channel_id is not None:
        
            if thresholds is not None:
                try:
                    if current_value >= thresholds:
                        msg = '[모델 학습 완료] ' + metric + '이 설정하신 thresolds: ' + str(thresholds) + '를 초과하였습니다.'
                        response = WebClient(token=token).chat_postMessage(channel=channel_id, text=msg)
                except SlackApiError as e:
                    logger.error(f"Error message: {e}")
                    
            elif message is not None:
                try:
                    msg = message
                    response = WebClient(token=token).chat_postMessage(channel=channel_id, text=msg)
                except SlackApiError as e:
                    logger.error(f"Error message: {e}")
        else:
            pass

    def _off_autolog():
        def stop_log(self, original, args, kwargs, unlogged_params):
            fit_model = original(self, *args, **kwargs)
            return (fit_model)

        @gorilla.patch(tensorflow.keras.Model)
        def fit(self, *args, **kwargs):
            original = gorilla.get_original_attribute(tensorflow.keras.Model, 'fit')
            unlogged_params = ['self', 'x', 'y', 'callbacks', 'validation_data', 'verbose']
            return stop_log(self, original, args, kwargs, unlogged_params)

        settings = gorilla.Settings(allow_hit=True, store_hit=True)
        gorilla.apply(gorilla.Patch(tensorflow.keras.Model, 'fit', fit, settings=settings))




# metrics in lightgbm

REGRESSION = ['l1', 'mean_absolute_error', 'mae', 'regression_l1',
             'l2', 'mean_squared_error', 'mse', 'regression_l2', 'regression',
             'rmse', 'root_mean_squared_error', 'l2_root',
             'mape', 'mean_absolute_percentage_error',
             'huber',
             'fair',
             'poisson',                    # Poisson regression
             'gamma', 'gamma_deviance',    # Gamma regression
             'tweedie',                    # Tweedie regression
             'quantile'                    # Quantile regression
             ]

CLASSIFICATION = ['map', 'mean_average_precision',
              'auc',
              'average_precision',
              'binary_logloss', 'binary',
              'binary_error',
              'auc_mu',
              'multi_logloss', 'multiclass', 'softmax',
              'multiclassova', 'multiclass_ova', 'ova', 'ovr',
              'multi_error',
              'cross_entropy', 'xentropy',
              'cross_entropy_lambda', 'xentlambda',
              'kullback_leibler', 'kldiv']
              
no_metrics = ['"None"', 'na', 'null', '']




"""
This http client send run's results data to modeler back-end server
"""

from Accuinsight.modeler.store.tracking.life_cycle_store import RestStore as LcRestStore
from Accuinsight.modeler.store.tracking.monitoring_deploy_store import RestStore as MonitoringDeployRestStore
from Accuinsight.modeler.store.tracking.workspace_run_store import RestStore as WorkspaceRunRestStore
from Accuinsight.modeler.utils.rest_utils import ModelerHostCreds
from Accuinsight.modeler.utils.os_getenv import get_os_env
from Accuinsight.modeler.core.LcConst import LcConst

_DEFAULT_USER_ID = "unknown"


class LifecycleRestApi:
    def __init__(self, host_url, port, uri):
        self.base_url = 'http://' + host_url + ':' + str(port) + '/' + uri

    def lc_create_run(self):
        """Create run."""

        store = LcRestStore(lambda: ModelerHostCreds(self.base_url))
        run = store.lc_create_run()
        return run


class DeployLogRestApi:
    def __init__(self, host_url, port, uri):
        self.base_url = 'http://' + host_url + ':' + str(port) + '/' + uri

    def call_rest_api(self, method, param, mode="logging"):
        store = MonitoringDeployRestStore(ModelerHostCreds(self.base_url))
        response = store.call_endpoint(method, param, mode)

        return response


class WorkspaceRunRestApi:
    def __init__(self, host_url, port, uri):
        self.base_url = 'http://' + host_url + ':' + str(port) + '/' + uri

    def call_rest_api(self, param):
        store = WorkspaceRunRestStore(ModelerHostCreds(self.base_url))
        response = store.call_endpoint(param)

        return response


if __name__ == "__main__":
    env_value = get_os_env('ENV')
    modeler_rest = LifecycleRestApi(env_value[LcConst.BACK_END_API_URL],
                                    env_value[LcConst.BACK_END_API_PORT],
                                    env_value[LcConst.BACK_END_API_URI])

    modeler_rest.lc_create_run()



import base64
import time
import logging
import json

import requests

from Accuinsight.__about__ import __version__
from Accuinsight.modeler.protos import skapi_pb2
from Accuinsight.modeler.utils.proto_json_utils import parse_dict
from Accuinsight.modeler.utils.string_utils import strip_suffix
from Accuinsight.modeler.exceptions import ModelerException, RestException
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import get_current_run
from Accuinsight.modeler.core.LcConst.LcConst import RUN_RESULT_PATH, RUN_MODEL_SHAP_JSON_PATH

RESOURCE_DOES_NOT_EXIST = 'RESOURCE_DOES_NOT_EXIST'

_logger = logging.getLogger(__name__)

_DEFAULT_HEADERS = {
    'User-Agent': 'modeler-python-client/%s' % __version__
}


def http_request(host_creds, endpoint, retries=3, retry_interval=3,
                 max_rate_limit_interval=60, **kwargs):
    """
    Makes an HTTP request with the specified method to the specified hostname/endpoint. Ratelimit
    error code (429) will be retried with an exponential back off (1, 2, 4, ... seconds) for at most
    `max_rate_limit_interval` seconds.  Internal errors (500s) will be retried up to `retries` times
    , waiting `retry_interval` seconds between successive retries. Parses the API response
    (assumed to be JSON) into a Python object and returns it.

    :param host_creds (object containing)
        hostname and optional authentication.
    :return: Parsed API response
    """
    hostname = host_creds.host
    auth_str = None
    if host_creds.username and host_creds.password:
        basic_auth_str = ("%s:%s" % (host_creds.username, host_creds.password)).encode("utf-8")
        auth_str = "Basic " + base64.standard_b64encode(basic_auth_str).decode("utf-8")
    elif host_creds.token:
        auth_str = "Bearer %s" % host_creds.token

    headers = dict(_DEFAULT_HEADERS)
    if auth_str:
        headers['Authorization'] = auth_str

    verify = not host_creds.ignore_tls_verification

    def request_with_ratelimit_retries(max_rate_limit_interval, **kwargs):
        response = requests.request(**kwargs)
        time_left = max_rate_limit_interval
        sleep = 1
        while response.status_code == 429 and time_left > 0:
            _logger.warning(
                "API request to {path} returned status code 429 (Rate limit exceeded). "
                "Retrying in %d seconds. "
                "Will continue to retry 429s for up to %d seconds.",
                sleep, time_left)
            time.sleep(sleep)
            time_left -= sleep
            response = requests.request(**kwargs)
            sleep = min(time_left, sleep*2)  # sleep for 1, 2, 4, ... seconds;
        return response

    cleaned_hostname = strip_suffix(hostname, '/')
    url = "%s%s" % (cleaned_hostname, endpoint)
    for i in range(retries):
        # response = request_with_ratelimit_retries(max_rate_limit_interval,
        #                                           url=url, headers=headers, verify=verify, **kwargs)
        response = request_with_ratelimit_retries(max_rate_limit_interval,
                                                  url=url, headers=headers, **kwargs)
        if response.status_code >= 200 and response.status_code < 500:
            return response
        else:
            _logger.error(
                "API request to %s failed with code %s != 200, retrying up to %s more times. "
                "API response body: %s",
                url, response.status_code, retries - i - 1, response.text)
            time.sleep(retry_interval)
    raise ModelerException("API request to %s failed to return code 200 after %s tries" %
                          (url, retries))


def _can_parse_as_json(string):
    try:
        json.loads(string)
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def http_request_safe(host_creds, endpoint, **kwargs):
    """
    Wrapper around ``http_request`` that also verifies that the request succeeds with code 200.
    """
    response = http_request(host_creds=host_creds, endpoint=endpoint, **kwargs)
    return verify_rest_response(response, endpoint)


def verify_rest_response(response, endpoint):
    """Verify the return code and raise exception if the request was not successful."""
    if response.status_code != 200:
        if _can_parse_as_json(response.text):
            raise RestException(json.loads(response.text))
        else:
            base_msg = "API request to endpoint %s failed with error code " \
                       "%s != 200" % (endpoint, response.status_code)
            raise ModelerException("%s. Response body: '%s'" % (base_msg, response.text))
    return response


def _get_path(path_prefix, endpoint_path):
    return "{}{}".format(path_prefix, endpoint_path)


def extract_api_info_for_service(service, path_prefix):
    """ Return a dictionary mapping each API method to a tuple (path, HTTP method)"""
    service_methods = service.DESCRIPTOR.methods
    res = {}
    for service_method in service_methods:
        endpoints = service_method.GetOptions().Extensions[skapi_pb2.rpc].endpoints
        endpoint = endpoints[0]
        endpoint_path = _get_path(path_prefix, endpoint.path)
        res[service().GetRequestClass(service_method)] = (endpoint_path, endpoint.method)
    return res


def _make_metrics_data(json_body):
    result = []
    # metrics_object = {'key': '', 'values': [], 'timestamp': [], 'steps': []}

    # get metrics
    metrics_data = json_body['metrics']
    # get visuals
    if 'visuals' in json_body:
        visuals_data = json_body['visuals']
        for item in visuals_data:
            # metrics_object = {'key': '', 'values': [], 'timestamp': [], 'steps': []}
            metrics_object = {'key': item['key'], 'values': [], 'timestamp': item['timestamp'], 'steps': item['steps']}
            # key, listValues, timestamp, steps

            if 'listValues' in item:
                for values in item['listValues'][0]['values']:
                    metrics_object['values'].append(values['values'])
            else:
                metrics_object['values'].extend(item['values'][0]['values'])

            result.append(metrics_object)
            metrics_data.append(metrics_object)

    if 'visuals' in json_body:
        del json_body['visuals']

    return json_body


def _make_parameter_data(json_body):
    result = []
    # metrics_object = {'key': '', 'values': [], 'timestamp': [], 'steps': []}

    # get metrics
    parameter_data = json_body['parameter']
    # "parameter": {
    #     "parameter": {
    #       "learningrate": "0.12",
    #       "subsample": "0.72",
    #       "colsamplebytree": "0.62",
    #       "maxdepth": "52",
    #       "alpha": "0.82"
    #     }
    #   }

    parameter = {'parameter': {}}
    for param in parameter_data:
        key = param['key']
        value = param['value']
        parameter['parameter'].setdefault(key, value)
        # column_dict.setdefault(col_name, []).append(row[col_name])

    if 'parameter' in json_body:
        del json_body['parameter']

    json_body['parameter'] = parameter
    return json_body

def _make_feature_data(json_body):
    json_body['feature'] = {}
    json_body['feature']['importance'] = {}

    run_info_json = get_current_run()
    run_result_path = run_info_json[RUN_RESULT_PATH]

    #json_path = run_result_path[RUN_MODEL_SHAP_JSON_PATH]

    #json_path = None
    
    if RUN_MODEL_SHAP_JSON_PATH in run_result_path.keys():
        json_path = run_result_path[RUN_MODEL_SHAP_JSON_PATH]
        
        with open(json_path) as json_file:
            shap_json_data = json.load(json_file)
        json_body['feature']['importance'] = shap_json_data   

    return json_body


def call_endpoint(host_creds, endpoint, method, json_body, response_proto):
    # Cover json String to json dictionary, to pass to requests
    if json_body:
        json_body = json.loads(json_body)
        json_body = _make_metrics_data(json_body)
        json_body = _make_parameter_data(json_body)
        json_body = _make_feature_data(json_body)
    if method == 'GET':
        response = http_request(
            host_creds=host_creds, endpoint=endpoint, method=method, params=json_body)
    else:
        response = http_request(
            host_creds=host_creds, endpoint=endpoint, method=method, json=json_body)
    response = verify_rest_response(response, endpoint)
    js_dict = json.loads(response.text)
    # print('Backend Response result: ', js_dict)
    parse_dict(js_dict=js_dict, message=response_proto)
    return response_proto


def key_clock(username, password):
  headers = {'Content-Type': 'application/x-www-form-urlencoded'}
  data = {'grant_type': 'password',
          'client_id': 'accu_modeler_client',
          'client_secret': '379668c7-70d2-49db-b977-d3c2b6baed14' ,
          'username': username,
          'password': password}
  try:
    response = requests.post('https://sso.aws-multiverse.skcc.com/auth/realms/DOTS/protocol/openid-connect/token', data, headers=headers)

  except Exception:
    logging.error("key_clock server connection failed")
    response = None

  return response

def daapi_list(host, api, username, password):
  try:
    token =  key_clock(usernamehashlib.py, password)
    headers = {'x-auth-token': token.json()["access_token"]}
    response = requests.get(host+api, headers=headers)
    result = response.json()["data"]

  except Exception:
    logging.error("DaApi server connection failed")
    result = None

  return result

def daapi_data_list(host, api, username, password, data):
  try:
    token =  key_clock(username, password)
    headers = {'x-auth-token': token.json()["access_token"]}
    response = requests.get(host+api, data, headers=headers)
    result = response.json()["data"]

  except Exception:
    logging.error("DaApi server connection failed")
    result = None

  return result


class ModelerHostCreds(object):
    """
    Provides a hostname and optional authentication for talking to an back-end server.
    :param host: Hostname (e.g., http://localhost:5000) to back-end server. Required.
    :param username: Username to use with Basic authentication when talking to server.
        If this is specified, password must also be specified.
    :param password: Password to use with Basic authentication when talking to server.
        If this is specified, username must also be specified.
    :param token: Token to use with Bearer authentication when talking to server.
        If provided, user/password authentication will be ignored.
    :param ignore_tls_verification: If true, we will not verify the server's hostname or TLS
        certificate. This is useful for certain testing situations, but should never be
        true in production.
    """
    def __init__(self, host, username=None, password=None, token=None,
                 ignore_tls_verification=False):
        if not host:
            raise ModelerException("host is a required parameter for ModelerHostCreds")
        self.host = host
        self.username = username
        self.password = password
        self.token = token
        self.ignore_tls_verification = ignore_tls_verification



'''
<2020.06.23>
Classify available models to regression and classification.

reference: https://scikit-learn.org/stable/index.html
'''

REGRESSION = ['Regressor', 'Ridge', 'Lasso', 'ElasticNet', 'Lars', 'SVR', 'OrthogonalMatchingPursuit']
CLASSIFICATION = ['Classifier', 'Logistic', 'SVC', 'NB', 'Perceptron']
ONEVS = ['OneVsRest', 'OneVsOne']


